#include <stdlib.h>
#include <stdio.h>
#include "joint.h"
#include "solide.h"
#include "solveur.h"



Joint::Joint(Solide *p1,Solide *p2,
             Vect3d const&bar1A1,Vect3d const&A2bar2):
     G1A1(bar1A1),G2A2(-A2bar2)
{
    libertes=3;
    part1 = p1;
    part2 = p2;

    (*p1).liens.add(this);
    (*p2).liens.add(this);
}




Vect3d Joint::A()const {return (*part2).p + (*part2).R*G2A2;}
Vect3d Joint::AG2()const {return -(*part2).R*G2A2;}
Vect3d Joint::AG1()const {return (*part1).p - ( (*part2).p + (*part2).R*G2A2 );}


void Joint::updatevar(){
    Vect3d &w1 = (*part1).w;
    Vect3d &w2 = (*part2).w;
    Vect3d &v1 = (*part1).v;
    Vect3d &v2 = (*part2).v;
    Vect3d &p1 = (*part1).p;
    Vect3d &p2 = (*part2).p;
    Matr9d &R1 = (*part1).R;
    Matr9d &R2 = (*part2).R;


    errc = p2 + R2*G2A2 - R1*G1A1 - p1;
    errv = v2 + (w2 ^ R2*G2A2) - (w1 ^ R1*G1A1) - v1;
    Gq1 = (R1*G1A1).v();
    Gp1 = Matr9d(-1);
    Gq2 = (R2*-G2A2).v();
    Gp2 = Matr9d(1);
    terme2 = (w1.v()*R1*G1A1).v()*w1 + (w2.v()*R2*-G2A2).v()*w2;

}


/*
On a une contrainte C(x)=0,
on veut la derivee deux fois de sorte à avoir:
grad(C(x))*x'' + grad(C(x))'*x' = 0
On a aussi d'après l'equation d'euler lagrange (elle nous donne les forces):
M x'' + M'x'= t(grad(C(x)))*lambda

L'equation que nous devons ecrire est alors:
grad(C(x))*(1/M)*t(grad(C(x)))*lambda = - grad(C(x))'*x' + grad(C(x))*(1/M)*M'x'

Pour la rotation on a: Iw' = t(grad(C(x)))*lambda - w^Iw
soit:
grad(C(x))*(1/I)*grad(C(x))*lambda = - grad(C(x))'*x' + grad(C(x))*(1/I)*(w^Iw)

Si nous voulons ecrire cette equation pour une contrainte seulement alors cela devient:
grad(Ci(x))*(1/I)*grad(C(x))*lambda = - grad(Ci(x))'*x' + grad(Ci(x))*(1/I)*(w^Iw)

Nos contraintes ne font intervenir que 4 variables q1,p1,q2,p2.
Il y aura donc quatre matrice 3*3 pour representer  grad(Ci(x)).
Ensuite il faut rajouter les gradient des contraintes qui font intervenir ces variables, pour le grad(C(x)) qui vient après.

*/


void Joint::corrigevar(){
    Solide &pere=*part1;
    Solide &fils=*part2;
    //correction en rotation (pas de correction)
    Matr9d P0;
    Vect3d diff;
    /*
    P0 = getbase()*getprojection(liberte+3)*transpo(getbase());

    fils.quat=(P0*(fils.quat/pere.quat))*pere.quat;
    fils.quat.renorme();
    fils.R = fils.quat.matr();

    Vect3d diff=fils.w - pere.w;
    fils.w -= diff-P0*diff;
    //*/

    //correction en déplacement
    P0 = Matr9d(0); //getbase()*getprojection(liberte)*transpo(getbase());
    diff = errc;
    fils.p -= diff-P0*diff;

    //Attention : cette correction comprend la dérivation de la base (G1A = G1A1+A1A2) en supposant que la base est celle de p1
    diff=errv;
    fils.v -= diff-P0*diff;
}



// liaison appuis-plan:
// projection de A2-A1 sur le plan dois être nul.
// produit vectoriel x1^x2 sur le plan doit être nul
void Appuiplan::updatevar(){
    Vect3d &w1 = (*part1).w;
    Vect3d &w2 = (*part2).w;
    Vect3d &v1 = (*part1).v;
    Vect3d &v2 = (*part2).v;
    Vect3d &p1 = (*part1).p;
    Vect3d &p2 = (*part2).p;
    Matr9d &R1 = (*part1).R;
    Matr9d &R2 = (*part2).R;


    Vect3d vx1=Vect3d(1,0,0);
    Vect3d vx2=Vect3d(1,0,0);

    errc = vx1 ^ transpo(R2)*R1*vx2;
    errc.x = (transpo(R2)*(p2 - R1*G1A1 - p1) + G2A2).x;

    errv = vx1 ^ (transpo(w2.v()*R2) + transpo(R2)*w1.v())*R1*vx2;
    errv.x = (transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1) + transpo(R2)*(v2 - (w1^R1*G1A1) - v1)).x;

    Gq1 = vx1.v()*transpo(R2)*(R1*-vx2).v();
    Gq1.Lx = (transpo(R2)*(R1*G1A1).v()).Lx;
    Gp1 = Matr9d(0);
    Gp1.Lx =  -transpo(R2).Lx;
    Gq2 = vx1.v()*transpo(R2)*(R1*vx2).v();
    Gq2.Lx = (transpo(R2)*(p2 - R1*G1A1 - p1).v()).Lx;
    Gp2 = Matr9d(0);
    Gp2.Lx =  transpo(R2).Lx;
    //*/

    terme2 = vx1.v()*(transpo(w2.v()*R2)*(R1*vx2).v()+ transpo(R2)*(w1.v()*R1*vx2).v() )*(w2-w1);
    terme2.x = transpo(w2.v()*R2).Lx*(v2-v1)
    + (transpo(w2.v()*R2)*(R1*G1A1).v()+ transpo(R2)*(w1.v()*R1*G1A1).v()).Lx*w1
    + (transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1).v() + transpo(R2)*(v2 - w1.v()*R1*G1A1 - v1).v()).Lx*w2;

}

// liaison lineaire annulaire:
// projection de A2-A1 sur le plan dois être nul.
void Annulaire::updatevar(){
    Vect3d &w1 = (*part1).w;
    Vect3d &w2 = (*part2).w;
    Vect3d &v1 = (*part1).v;
    Vect3d &v2 = (*part2).v;
    Vect3d &p1 = (*part1).p;
    Vect3d &p2 = (*part2).p;
    Matr9d &R1 = (*part1).R;
    Matr9d &R2 = (*part2).R;


    errc = transpo(R2)*(p2 - R1*G1A1 - p1) + G2A2;
    errc.z = 0;

    errv = transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1) + transpo(R2)*(v2 - (w1^R1*G1A1) - v1);
    errv.z = 0;

    Gq1 = transpo(R2)*(R1*G1A1).v();
    Gq1.Lz = Vect3d();
    Gp1 = -transpo(R2);
    Gp1.Lz =  Vect3d();
    Gq2 = transpo(R2)*(p2 - R1*G1A1 - p1).v();
    Gq2.Lz = Vect3d();
    Gp2 = transpo(R2);
    Gp2.Lz =  Vect3d();

    terme2 = transpo(w2.v()*R2)*(v2-v1)
    + (transpo(w2.v()*R2)*(R1*G1A1).v()+ transpo(R2)*(w1.v()*R1*G1A1).v())*w1
    + (transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1).v() + transpo(R2)*(v2 - w1.v()*R1*G1A1 - v1).v())*w2;
    terme2.z=0;

}

// liaison lineaire annulaire:
// projection de A2-A1 sur le plan dois être nul.
void Ponctuelle::updatevar(){
    Vect3d &w1 = (*part1).w;
    Vect3d &w2 = (*part2).w;
    Vect3d &v1 = (*part1).v;
    Vect3d &v2 = (*part2).v;
    Vect3d &p1 = (*part1).p;
    Vect3d &p2 = (*part2).p;
    Matr9d &R1 = (*part1).R;
    Matr9d &R2 = (*part2).R;


    errc = transpo(R2)*(p2 - R1*G1A1 - p1) + G2A2;
    errc.z = errc.y = 0;

    errv = transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1) + transpo(R2)*(v2 - (w1^R1*G1A1) - v1);
    errv.z =errv.y = 0;

    Gq1 = transpo(R2)*(R1*G1A1).v();
    Gq1.Lz = Gq1.Ly = Vect3d();
    Gp1 = -transpo(R2);
    Gp1.Lz = Gp1.Ly = Vect3d();
    Gq2 = transpo(R2)*(p2 - R1*G1A1 - p1).v();
    Gq2.Lz = Gq2.Ly = Vect3d();
    Gp2 = transpo(R2);
    Gp2.Lz = Gp2.Ly = Vect3d();

    terme2 = transpo(w2.v()*R2)*(v2-v1)
    + (transpo(w2.v()*R2)*(R1*G1A1).v()+ transpo(R2)*(w1.v()*R1*G1A1).v())*w1
    + (transpo(w2.v()*R2)*(p2 - R1*G1A1 - p1).v() + transpo(R2)*(v2 - w1.v()*R1*G1A1 - v1).v())*w2;
    terme2.z = terme2.y = 0;

}

void Joint::fillequa(Matr9d *equa){

    int j;
    Solide &p1 = *part1;
    Solide &p2 = *part2;
    Matr9d I1 = p1.getinertie();
    Matr9d I2 = p2.getinertie();


	//contributions p1
    for(j=0;j<p1.liens.end();j++){
        Joint &p1j = *p1.liens(j);
        if(p1j.is_pere(p1)){
            equa[p1j.iforce] += Gq1*!I1*transpo(p1j.Gq1) + 1./p1.masse*Gp1*transpo(p1j.Gp1);
        }else{
            equa[p1j.iforce] += Gq1*!I1*transpo(p1j.Gq2) + 1./p1.masse*Gp1*transpo(p1j.Gp2);
        }
    }

    //contributions p2
    for(j=0;j<p2.liens.end();j++){
        Joint &p2j = *p2.liens(j);
        if(p2j.is_pere(p2)){
            equa[p2j.iforce] += Gq2*!I2*transpo(p2j.Gq1) + 1./p2.masse*Gp2*transpo(p2j.Gp1);
        }else{
            equa[p2j.iforce] += Gq2*!I2*transpo(p2j.Gq2) + 1./p2.masse*Gp2*transpo(p2j.Gp2);
        }
    }
    if(libertes==2)equa[iforce]+=Matr9d(0,0,1);
    else if(libertes==1)equa[iforce]+=Matr9d(0,1,1);
}







