#ifndef JOINT_H
#define JOINT_H

#include "matr9d.h"
#include "patrons.h"


class Solide;
class Joint;

// L'ensemble des liaisons modélisées (générateur) est : rotule ; lineaire annulaire ; ponctuel ; appui plan ; cylindre plan.

class Joint{
    public:
    Joint(Solide *p1,Solide *p2,Vect3d const&bar1A1,Vect3d const&A2bar2);
    inline bool is_pere(Solide &s){return &s==part1;}

    // précalcul les gradients, les erreur de contrainte, et le second terme
    virtual void updatevar();
    // remplissage de l'équation associée à la contrainte
    void fillequa(Matr9d *equa);

	Vect3d A()const; // donne la position d'attache du joint, utile pour l'affichage
    Vect3d AG2()const; // donne la position d'attache du joint par rapport au solide 2
    Vect3d AG1()const; // donne la position d'attache du joint par rapport au solide 1
    void corrigevar(); // corrige la contrainte pour un système arborescent

    //variables de stockage
    Vect3d errc; // valeur de la fonction contrainte (doit valoir 0)
    Vect3d errv; // dérivée de la fonction contrainte (doit valoir 0)
    Vect3d terme2; // second terme: J'u' qui apparait dans l'équation
    Matr9d Gq1,Gq2,Gp1,Gp2; // les gradients

    //caracteristiques de la liaison (constantes)
    Matr9d base0; // si la contrainte s'exprime dans une certaine base
    Quaternion quat0;

	int iforce; // numéro de la force associé au joint
    int libertes; // nombre de liberté (entre 0 et 3)

    Solide* part1; // pointeur vers le solide 1
    Solide* part2; // pointeur vers le solide 2
    Vect3d G1A1; // position de la liaison dans la base locale au solide 1
    Vect3d G2A2; // position de la liaison dans la base locale au solide 2
};

class Appuiplan : public Joint{
    public:
    virtual void updatevar();

    inline Appuiplan(Solide *p1,Solide *p2, Vect3d const&bar1A1,Vect3d const&A2bar2):Joint(p1,p2,bar1A1,A2bar2){}

};

class Annulaire : public Joint{
    public:
    virtual void updatevar();

    inline Annulaire(Solide *p1,Solide *p2, Vect3d const&bar1A1,Vect3d const&A2bar2):Joint(p1,p2,bar1A1,A2bar2){libertes=2;}

};

class Ponctuelle : public Joint{
    public:
    virtual void updatevar();

    inline Ponctuelle(Solide *p1,Solide *p2, Vect3d const&bar1A1,Vect3d const&A2bar2):Joint(p1,p2,bar1A1,A2bar2){libertes=1;}

};






#endif
