#ifndef MATR9D_H
#define MATR9D_H

#include "patrons.h"
#include "stdio.h"
#include <math.h>
#include <GL/glew.h>


class Matr9d;



class Vect3d{
    public:

    inline Vect3d(){x=y=z=0;}
    inline Vect3d(double a,double b, double c){x=a;y=b;z=c;}

    inline double operator[](int i)const{return ((double*)this)[i];}
    inline double& operator[](int i){return ((double*)this)[i];}


    inline Matr9d v() const;//renvoie la matrice du produit vectoriel correspondant
    inline void print (const char* nom) const{printf("%s : %f %f %f \n",nom,x,y,z);}

    inline void operator+=(Vect3d const&a){x+=a.x;y+=a.y;z+=a.z;}
    inline void operator-=(Vect3d const&a){x-=a.x;y-=a.y;z-=a.z;}
    inline void operator*=(double a){x*=a;y*=a;z*=a;}

    //pour le pivot de Gauss
    inline void ajoute(Matr9d const&m, Vect3d const&v);
    inline void fois(Matr9d const&m);



    double x;
    double y;
    double z;
};

inline Vect3d operator-(Vect3d const& a){return Vect3d(-a.x,-a.y,-a.z);}

inline Vect3d operator+( Vect3d const&a,Vect3d const&b){Vect3d c(a);c+=b;return c;}
inline Vect3d operator-( Vect3d const&a,Vect3d const&b){Vect3d c(a);c-=b;return c;}

inline Vect3d operator*(double a,Vect3d const& b){Vect3d v(b);v*=a;return v;}
inline double operator*(Vect3d const& a,Vect3d const& b){return a.x*b.x+a.y*b.y+a.z*b.z;}

inline Vect3d vectorielle(Vect3d const& a,Vect3d const& b){
    return Vect3d(
        a.y*b.z-a.z*b.y,
        b.x*a.z-b.z*a.x,
        a.x*b.y-a.y*b.x);
}
inline Vect3d operator^(Vect3d const& a,Vect3d const& b){
    return Vect3d(
        a.y*b.z-a.z*b.y,
        b.x*a.z-b.z*a.x,
        a.x*b.y-a.y*b.x);
}


class Matr9d{
    public:
    inline Matr9d():Lx(),Ly(),Lz(){}
    inline Matr9d(double d):Lx(d,0,0),Ly(0,d,0),Lz(0,0,d){}
    inline Matr9d(double a,double b, double c):Lx(a,0,0),Ly(0,b,0),Lz(0,0,c){}
    inline Matr9d(Vect3d const& diagj):Lx(diagj.x,0,0),Ly(0,diagj.y,0),Lz(0,0,diagj.z){}
    inline Matr9d(Vect3d const&l1, Vect3d const&l2, Vect3d const&l3){Lx=l1;Ly=l2;Lz=l3;}
    inline double operator[](int i)const{return ((double*)this)[i];}
    inline double& operator[](int i){return ((double*)this)[i];}

    inline void print (const char* nom) const{printf("%s\n",nom);Lx.print("");Ly.print("");Lz.print("");}




    inline void transpo(){
        echange(Lx.y,Ly.x);
        echange(Lz.y,Ly.z);
        echange(Lz.x,Lx.z);
    }
    void inline operator+=(Matr9d const& a){Lx+=a.Lx;Ly+=a.Ly;Lz+=a.Lz;}
    void inline operator-=(Matr9d const& a){Lx-=a.Lx;Ly-=a.Ly;Lz-=a.Lz;}
    void inline operator*=(double a){Lx*=a;Ly*=a;Lz*=a;}

    //pour le pivot de Gauss
    inline void ajoute(Matr9d const&m, Matr9d const&v);
    inline void fois(Matr9d const&m);

    Vect3d Lx;
    Vect3d Ly;
    Vect3d Lz;

    //static long nmul;
};




inline Matr9d transpo(Matr9d const& m){Matr9d a(m);a.transpo();return a;}
inline Matr9d operator-(Matr9d const& a){return Matr9d(-a.Lx,-a.Ly,-a.Lz);}

inline Matr9d operator+( Matr9d const&a,Matr9d const&b){Matr9d c(a);c+=b;return c;}
inline Matr9d operator-( Matr9d const&a,Matr9d const&b){Matr9d c(a);c-=b;return c;}


inline Vect3d operator*(Matr9d const& m, Vect3d const& a){
    return Vect3d(m.Lx*a,m.Ly*a,m.Lz*a);
}
inline Matr9d operator*(double a, Matr9d const& b){
    Matr9d c(b); c*=a; return c;}

inline Matr9d operator*(Matr9d const& a, Matr9d const& b){
    //Matr9d::nmul++;
    return Matr9d(a.Lx.x*b.Lx+a.Lx.y*b.Ly+a.Lx.z*b.Lz,
                  a.Ly.x*b.Lx+a.Ly.y*b.Ly+a.Ly.z*b.Lz,
                  a.Lz.x*b.Lx+a.Lz.y*b.Ly+a.Lz.z*b.Lz);
}

inline Matr9d operator!(Matr9d const& b){
    Matr9d m;
    m[0]=b[4]*b[8]-b[5]*b[7];
    m[1]=b[2]*b[7]-b[8]*b[1];
    m[2]=b[1]*b[5]-b[2]*b[4];
    m[3]=b[5]*b[6]-b[8]*b[3];
    m[4]=b[0]*b[8]-b[6]*b[2];
    m[5]=b[2]*b[3]-b[0]*b[5];
    m[6]=b[3]*b[7]-b[4]*b[6];
    m[7]=b[1]*b[6]-b[0]*b[7];
    m[8]=b[0]*b[4]-b[1]*b[3];
    //if(fabs(vectorielle(b.Lx,b.Ly)*b.Lz)<0.000000001)b.print("");
    return (1./(vectorielle(b.Lx,b.Ly)*b.Lz))*m;
}
//*
inline Matr9d pivotaxeangle(Vect3d const& axe, double teta){
    Matr9d m(axe.v());
    m = sin(teta)*m+(1-cos(teta))*m*m;
    m[0]+=1;
    m[4]+=1;
    m[8]+=1;
    return m;
}

inline Matr9d dpivotaxe(Vect3d const& axe){
    Matr9d m;
    m.Lx = Vect3d(1,-axe.z,axe.y);
    m.Ly = Vect3d(axe.z,1,-axe.x);
    m.Lz = Vect3d(-axe.y,axe.x,1);
    return m;
}
inline Matr9d dpivotaxe2(Vect3d const& axe){
    Matr9d m(Vect3d(0,-axe.z,axe.y),
              Vect3d(axe.z,0,-axe.x),
              Vect3d(-axe.y,axe.x,0));
    m += 0.5*m*m;
    m.Lx.x+=1;
    m.Ly.y+=1;
    m.Lz.z+=1;
    return m;
}
Matr9d getprojection(const bool *tab3);

// pour le pivot de Gauss
inline void Vect3d::ajoute(Matr9d const&m, Vect3d const&v){*this += m*v;}
inline void Vect3d::fois(Matr9d const&m){*this = m*(*this);}

inline void Matr9d::ajoute(Matr9d const&m, Matr9d const&v){*this += m*v;}
inline void Matr9d::fois(Matr9d const&m){*this = m*(*this);}


class Quaternion{
    public:

    inline Quaternion():v(){s=1;}//quaternion unitaire
    inline Quaternion(double a,Vect3d const&u){s=a;v=u;}

    inline double operator[](int i)const{return ((double*)this)[i];}
    inline double& operator[](int i){return ((double*)this)[i];}

    inline void operator+=(Quaternion const& b){s+=b.s;v+=b.v;}
    inline void renorme(){
        double d=(s*s+v*v)*(-0.5)+1.5;
        s*=d;
        v*=d;
    }
    inline void renorme2(){
        double d0 = (s*s+v*v)-1;
        double d = 1+d0*(-0.5+d0*(0.375-d0*0.3125));//developpement au 3eme ordre 1-1/2*x+3/8*x²-5/16*x³
        s*=d;
        v*=d;
    }
    inline Matr9d matr()const{
        double ab = 2.*THIS[0]*THIS[1],
            ac= 2.*THIS[0]*THIS[2],ad= 2.*THIS[0]*THIS[3],
            bc= 2.*THIS[1]*THIS[2],bd= 2.*THIS[1]*THIS[3],
            cd= 2.*THIS[2]*THIS[3],bb= 2.*THIS[1]*THIS[1],
            cc= 2.*THIS[2]*THIS[2],dd= 2.*THIS[3]*THIS[3];
            return Matr9d(Vect3d(1-cc-dd,bc-ad,ac+bd),
                   Vect3d(ad+bc,1-bb-dd,cd-ab),
                   Vect3d(bd-ac,ab+cd,1-bb-cc));
    }

    double s;
    Vect3d v;
};
inline Quaternion operator+( Quaternion const&a,Quaternion const&b){Quaternion c(a);c+=b;return c;}

inline Quaternion renorme(Quaternion const& a){Quaternion b(a); b.renorme();return b;}
inline Quaternion dquat2(Vect3d const&a){return Quaternion(1-0.125*a*a,0.5*a);}
inline Quaternion operator*(Quaternion const&a, Quaternion const&b){return Quaternion(a.s*b.s-a.v*b.v, a.s*b.v+b.s*a.v+vectorielle(a.v,b.v));}
inline Quaternion operator*(Vect3d const&a, Quaternion const&b){return Quaternion(-a*b.v, b.s*a+vectorielle(a,b.v));}
inline Quaternion operator/(Quaternion const&a, Quaternion const&b){return Quaternion(a.s*b.s+a.v*b.v, b.s*a.v-a.s*b.v-vectorielle(a.v,b.v));}//même chose que operator* mais avec -b.v
inline Matr9d getprojection( const bool *tab3){return Matr9d((double)tab3[0],(double)tab3[1],(double)tab3[2]);}
inline Quaternion operator*(double a, Quaternion const& b){return Quaternion(a*b.s,a*b.v);}
inline Quaternion operator!(Quaternion const& a){return Quaternion(a.s,-a.v);}

inline Quaternion operator*(Matr9d const&a, Quaternion const&b){Quaternion c(b);c.v=a*c.v; return c;}

/*
inline void multglmatrix(Matr9d const& m){
    double m2[16] = //codage en colonne obligatoire...
    {
        m[0],m[3],m[6],0,
        m[1],m[4],m[7],0,
        m[2],m[5],m[8],0,
        0,0,0,1
    };
    glMultMatrixd(m2); //appel de la multiplication GL
}
inline void multgltmatrix(Matr9d const& m){
    double m2[16] = //codage en ligne car on transpose...
    {
        m[0],m[1],m[2],0,
        m[3],m[4],m[5],0,
        m[6],m[7],m[8],0,
        0,0,0,1
    };
    glMultMatrixd(m2); //appel de la multiplication GL
}//*/

inline Matr9d gramschmidt(Vect3d const&a){
    Matr9d m;
    m.Lx = (1/sqrt(a*a))*a;

    if(fabs(m.Lx.x)<fabs(m.Lx.y) && fabs(m.Lx.x)<fabs(m.Lx.z))m.Ly=Vect3d(1,0,0);
    else if(fabs(m.Lx.y)<fabs(m.Lx.z))m.Ly=Vect3d(0,1,0);
    else m.Ly=Vect3d(0,0,1);//on prend la plus petite valeur

    m.Ly -= (m.Ly*m.Lx)*m.Lx;
    m.Ly = (1/sqrt(m.Ly*m.Ly))*m.Ly;
    m.Lz = m.Lx^m.Ly;
    return m;
}

inline Matr9d Vect3d::v()const{
    return Matr9d(Vect3d(0,-z,y),
                  Vect3d(z,0,-x),
                  Vect3d(-y,x,0));
}




#endif
