#ifndef SOLVEUR_H
#define SOLVEUR_H

#include <GL/glew.h>
#include "matr9d.h"
#include "solide.h"
#include "joint.h"



#define DELAY_MS        20
#define DELAY_FRAME     20 // 50fps
#define PAS_TEMPS       (double)DELAY_MS/1000.0


/*
références :
http://www.inrialpes.fr/sed/people/boissieux/HUMANS/H-ANIM/

Les proportions humaines:
http://villemin.gerard.free.fr/Biologie/CorpsPro.htm

*/

#define PROP_TETE           0.069
#define PROP_TRONC          0.275
#define PROP_BASSIN         0.160

#define PROP_BRAS           0.027
#define PROP_AVANT_BRAS     0.016
#define PROP_MAIN           0.006
#define PROP_CUISSE         0.142
#define PROP_JAMBE          0.043
#define PROP_PIED           0.014

// pivot de gauss
void gauss(dynamictab<Matr9d> &s, dynamictab<Vect3d> &v);

class Solveur
{
    public:
    inline Solveur(){}
    Solveur(double poids); // constructeur
    ~Solveur(); // destructeur

    void accelere_rot(Vect3d const& dw); // permet de mettre en rotation tout le bonhomme

    int add(Solide *s);// ajoute un solide
    int link(unsigned int p1,unsigned int p2, Vect3d const&d1, Vect3d const&d2);// ajoute une liaison rotule


    void solvesystemeforce(); // calcul les multiplicateurs de Lagrange et les accélérations (forces)
    void incrementeforce(double dt); // Euler
    void incrementeforce2(double dt); // RK2
    void incrementeforce4(double dt); // RK4


    void corrigevars();// corrige pour des liaisons rotule arborescente
    double corrigepos();// correction en position (une itération de la méthode de Newton), retourne l'erreur
    void corrigev();// corrige en vitesse
    void impulsion(); // génère une impulsion si le solveur sort de la zone
    void corrige();// corrige la position et la vitesse en effectuant plusieurs itérations (méthode de Newton)

    double geterreur(); // retourne la norme cartésienne de la contrainte en position
    void inversedyn(); // dynamique inverse
    double getenergie(); // retourne l'energie totale du système
    void recentre(); // positionne le bonhomme au centre et supprime sa vitesse de groupe
    void affiche(); // affiche les segments

    // quantité totales du système
    double massetot;
    Vect3d barycentretot;
    Vect3d vtot;

    // liste des joints et solides
    dynamictab<Solide*> listepartie;
    dynamictab<Joint*> listejoints;

    // système à inverser préaloué
    dynamictab<Vect3d> constante;
    dynamictab<Matr9d> sys;
};

#endif
