#include <GL/glew.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "matr9d.h"
#include "solveur.h"

/// x : horizontal,    y : vertical,  z : profondeur



Solveur::Solveur(double poids){
    //Création des parties du corps.

    int pelvis = add(new Solide(PROP_BASSIN*poids,Vect3d(0.15,0.10,0.10)));
    int torso = add(new Solide(PROP_TRONC*poids,Vect3d(0.15,0.15,0.10)));
    int skull = add(new Solide(PROP_TETE*poids,Vect3d(0.08,0.10,0.08)));

    int r_upperarm = add(new Solide(PROP_BRAS*poids,Vect3d(0.04,0.15,0.04)));
    int r_forearm = add(new Solide(PROP_AVANT_BRAS*poids,Vect3d(0.03,0.15,0.03)));
    int r_hand = add(new Solide(PROP_MAIN*poids,Vect3d(0.01,0.08,0.05)));

    int l_upperarm = add(new Solide(PROP_BRAS*poids,Vect3d(0.04,0.15,0.04)));
    int l_forearm = add(new Solide(PROP_AVANT_BRAS*poids,Vect3d(0.03,0.15,0.03)));
    int l_hand = add(new Solide(PROP_MAIN*poids,Vect3d(0.01,0.08,0.05)));

    int r_thigh = add(new Solide(PROP_CUISSE*poids,Vect3d(0.05,0.20,0.05)));
    int r_calf = add(new Solide(PROP_JAMBE*poids,Vect3d(0.04,0.20,0.04)));
    int r_hindfoot = add(new Solide(PROP_PIED*poids,Vect3d(0.03,0.10,0.03)));

    int l_thigh = add(new Solide(PROP_CUISSE*poids,Vect3d(0.05,0.20,0.05)));
    int l_calf = add(new Solide(PROP_JAMBE*poids,Vect3d(0.04,0.20,0.04)));
    int l_hindfoot = add(new Solide(PROP_PIED*poids,Vect3d(0.03,0.10,0.03)));


    link(pelvis,torso,Vect3d(0,0.10,0),Vect3d(0,0.15,0.));
    link(torso,skull,Vect3d(0,0.15,0.10),Vect3d(0,0.12,0));

    link(torso,r_upperarm,Vect3d(0.12,0.15,0),Vect3d(0,-0.15,0));
    link(r_upperarm,r_forearm,Vect3d(0,-0.15,0),Vect3d(0,-0.15,0));
    link(r_forearm,r_hand,Vect3d(0,-0.15,0),Vect3d(0,-0.07,0));

    link(torso,l_upperarm,Vect3d(-0.12,0.15,0),Vect3d(0,-0.15,0));
    link(l_upperarm,l_forearm,Vect3d(0,-0.15,0),Vect3d(0,-0.15,0));
    link(l_forearm,l_hand,Vect3d(0,-0.15,0),Vect3d(0,-0.07,0));

    link(pelvis,l_thigh,Vect3d(-0.10,-0.10,0.10),Vect3d(0,-0.20,0));
    link(l_thigh,l_calf,Vect3d(0,-0.20,0),Vect3d(0,-0.20,0));
    link(l_calf,l_hindfoot,Vect3d(0,-0.20,0),Vect3d(0,-0.05,0.05));

    link(pelvis,r_thigh,Vect3d(0.10,-0.10,0.10),Vect3d(0,-0.20,0));
    link(r_thigh,r_calf,Vect3d(0,-0.20,0),Vect3d(0,-0.20,0));
    link(r_calf,r_hindfoot,Vect3d(0,-0.20,0),Vect3d(0,-0.05,0.05));



    //link(l_hand,r_hindfoot,Vect3d(0.,0,0),Vect3d(0.,0,-0));

}

Solveur::~Solveur(){
    int k;
    for(k=0;k<listepartie.end();k++) delete listepartie(k);
    for(k=0;k<listejoints.end();k++) delete listejoints(k);
}

int Solveur::add(Solide *s){
    listepartie.add(s);
    massetot+=(*s).masse;
    (*s).numero = listepartie.end()-1;
    return listepartie.end()-1;
}

int Solveur::link(unsigned int p1,unsigned int p2, Vect3d const&d1, Vect3d const&d2){
    if(p1!=p2 && p1 < listepartie.end() && p2 < listepartie.end()){
        listejoints.add(new Joint(listepartie(p1),listepartie(p2),d1,d2));
        (*listejoints(listejoints.end()-1)).iforce = listejoints.end()-1;
        return listejoints.end()-1;
    }else{
        return -1;
    }
}

void Solveur::accelere_rot(Vect3d const&dw){
    int k;
    for(k=0;k<listepartie.end();k++){(*listepartie(k)).w+=dw;}
}


void Solveur::affiche(){
    glBegin(GL_LINES);
    int k,j;
    Vect3d v;
    for(k=0;k<listepartie.end();k++){
        Solide &pk = *listepartie(k);
        for(j=1;j<pk.liens.end();j++){
            Joint &pj = *pk.liens(j);
            v = pk.A();
            glVertex3dv((GLdouble*)&v);
            v = pj.A();
            glVertex3dv((GLdouble*)&v);

            if((*pj.part2).liens.end()==1){//les extremités
                v = pj.A();
                glVertex3dv((GLdouble*)&v);
                v = v + 2*pj.AG2();
                glVertex3dv((GLdouble*)&v);
            }
        }
    }
    glEnd();
}


void Solveur::recentre(){
    Vect3d bar;
    Vect3d vit;
    int k;
    //calcul des vtot et bar
    for(k=0;k<listepartie.end();k++){
        Solide &pk = *listepartie(k);
        vit+=pk.masse*pk.v;
        bar+=pk.masse*pk.p;
    }
    vit*=1./massetot;
    bar*=1./massetot;

    for(k=0;k<listepartie.end();k++){
        Solide &pk = *listepartie(k);
        pk.v -= vit;//replacement en coordonnées barycentriques
        pk.p -= bar;
    }
    barycentretot+=bar;
    vtot+=vit;
}

void Solveur::corrige(){
    printf("erreurs %le\n", geterreur());
    int k=0;
    while(corrigepos()>0.000000000001 && k<100){// correction initiale
        k++;
        printf("erreurs %le\n", geterreur());
    }
    corrigev();// La corection en v est directe
    printf("convergence après %d itérations\n",k);
}



void gauss(dynamictab<Matr9d> &s, dynamictab<Vect3d> &v){
    int dim = v.end();
    if(dim*dim !=s.end()){
        printf("inversion du système de tailles incompatible dim^2 :  %d %d", v.end()*v.end(), s.end());exit(0);
    }
    // **************descente************

    int i,j,k;
	for(i=0; i<dim; i++){
	    Matr9d pivot = !s(i + dim*i);

	    // division de la ligne par le pivot
	    s(i + dim*i) = Matr9d();
        v(i) = pivot*v(i);
		for(j=i+1;j<dim;j++){
            s(j + dim*i) = pivot*s(j + dim*i);
        }
        // élimination des termes sous le pivot
        for(j=i+1;j<dim;j++){
            Matr9d coef = s(i+dim*j);
            s(i+dim*j) = Matr9d();
            v(j) -= coef*v(i);
            for(k=i+1;k<dim;k++){
                s(k + dim*j) -= coef*s(k + dim*i);
            }
        }
    }
    // ***********remontée***************
	for(i = dim-1; i>=0; i--){
        for(k=0; k<i; k++){
            v(k) -= s(i + dim*k)*v(i);
            s(i + dim*k) = Matr9d();
        }
    }
    // à la fin du pivot toute la matrice est nulle
}



