#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "matr9d.h"
#include "solveur.h"

extern int iteration;

void Solveur::solvesystemeforce(){
    int k,i;

    // initialisations
    constante.setend(listejoints.end());
    for(k=0;k<listejoints.end();k++){constante(k)=Vect3d(); }
    sys.setend(listejoints.end()*listejoints.end());
    for(k=0;k<sys.end();k++)sys(k)=Matr9d();

    for(k=0;k<listejoints.end();k++){// précalculs des gradients
        (*listejoints(k)).updatevar();
    }
    for(k=0;k<listejoints.end();k++){// remplissage du système
        Joint &j = *listejoints(k);
        j.fillequa(&sys(listejoints.end()*k));
        constante(k) -= j.terme2;// les termes constants s'ajoutent positivement à droite de l'equation
        constante(k) += j.Gq1*!(j.part1->getinertie())*(j.part1->w ^ (j.part1->getinertie() * j.part1->w));
        constante(k) += j.Gq2*!(j.part2->getinertie())*(j.part2->w ^ (j.part2->getinertie() * j.part2->w));
    }


    gauss(sys,constante);


    for(k=0;k<listepartie.end();k++){// calcul de l'accélération
        Solide &a = *listepartie(k);
        a.a_rot=-a.w^a.getinertie()*a.w;
        a.acceleration=Vect3d();

        for(i=0;i<a.liens.end();i++){
            Joint &j = *a.liens(i);
            if(j.is_pere(a)){
                a.a_rot += transpo(j.Gq1)*constante(j.iforce);
                a.acceleration += transpo(j.Gp1)*constante(j.iforce);
            }else{
                a.a_rot += transpo(j.Gq2)*constante(j.iforce);
                a.acceleration += transpo(j.Gp2)*constante(j.iforce);
            }
        }

        a.a_rot=!a.getinertie()*a.a_rot;
        a.acceleration*=1./a.masse;
    }


}

void Solveur::incrementeforce(double dt){
    int k;
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        a.q += dt*0.5*a.w*a.q;
        a.q.renorme();
        a.R = a.q.matr();
        a.w += dt*a.a_rot;

        a.p += dt*a.v;
        a.v += dt*a.acceleration;
    }
}
//*
void Solveur::incrementeforce2(double dt){
    int k;
    Quaternion dquat;
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        a.w0 = a.w;//on range l'iteration initiale
        a.q0 = a.q;
        a.p0 = a.p;
        a.v0 = a.v;
        dquat = 0.5*a.w*a.q;
        a.w = a.w0 + 0.5*dt*a.a_rot;//On prepare le calcul de k2
        a.q = a.q0 + 0.5*dt*dquat;
        a.p = a.p0 + 0.5*dt*a.v;
        a.v = a.v0 + 0.5*dt*a.acceleration;
        a.q.renorme();
        a.R = a.q.matr();
    }
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        dquat = 0.5*a.w*a.q;
        a.w = a.w0 + dt*a.a_rot;//On calcul l'itération
        a.q = renorme(a.q0 + dt*dquat);
        a.p = a.p0 + dt*a.v;
        a.v = a.v0 + dt*a.acceleration;
        a.R = a.q.matr();
    }
}
//*/
void Solveur::incrementeforce4(double dt){
    int k;
    Quaternion dquat;
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        a.w0 = a.w;//on range l'iteration initiale
        a.q0 = a.q;
        a.p0 = a.p;
        a.v0 = a.v;
        dquat = 0.5*a.w*a.q;
        a.q1 = dquat;//addition de k1 dans le résultat final
        a.w1 = a.a_rot;
        a.v1 = a.acceleration;
        a.p1 = a.v;
        a.w = a.w0 + 0.5*dt*a.a_rot;//On prepare le calcul de k2
        a.q = a.q0 + 0.5*dt*dquat;
        a.p = a.p0 + 0.5*dt*a.v;
        a.v = a.v0 + 0.5*dt*a.acceleration;
        a.q.renorme2();
        a.R = a.q.matr();
    }
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        dquat = 0.5*a.w*a.q;
        a.q1 += 2*dquat;//addition de k2 dans le résultat final
        a.w1 += 2*a.a_rot;
        a.v1 += 2*a.acceleration;
        a.p1 += 2*a.v;
        a.w = a.w0 + 0.5*dt*a.a_rot;//On prepare le calcul de k3
        a.q = a.q0 + 0.5*dt*dquat;
        a.p = a.p0 + 0.5*dt*a.v;
        a.v = a.v0 + 0.5*dt*a.acceleration;
        a.q.renorme2();
        a.R = a.q.matr();
    }
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        dquat = 0.5*a.w*a.q;
        a.q1 += 2*dquat;//addition de k3 dans le résultat final
        a.w1 += 2*a.a_rot;
        a.v1 += 2*a.acceleration;
        a.p1 += 2*a.v;
        a.w = a.w0 + dt*a.a_rot;//On prepare le calcul de k4
        a.q = a.q0 + dt*dquat;
        a.p = a.p0 + dt*a.v;
        a.v = a.v0 + dt*a.acceleration;
        a.q.renorme2();
        a.R = a.q.matr();
    }
    solvesystemeforce();
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        dquat = 0.5*a.w*a.q;
        a.q1 += dquat;//addition de k3 dans le résultat final
        a.w1 += a.a_rot;
        a.v1 += a.acceleration;
        a.p1 += a.v;
        a.w = a.w0 + dt/6.0*a.w1;//On calcule l'itération
        a.q = a.q0 + dt/6.0*a.q1;
        a.p = a.p0 +dt/6.0*a.p1;
        a.v = a.v0 + dt/6.0*a.v1;
        a.q.renorme2();
        a.R = a.q.matr();
    }
}



//*
void Solveur::inversedyn(){
    int k,i;
    Matr9d *equa1; // equations courante
    Matr9d *equa2;
    Vect3d *secondterme; // seconds terme equation courante

    // initialisations
    constante.setend(listejoints.end()*2);
    for(k=0;k<listejoints.end();k++){constante(k)=Vect3d();}
    sys.setend(listejoints.end()*listejoints.end()*4);
    for(k=0;k<sys.end();k++)sys(k)=Matr9d();

    for(k=0;k<listejoints.end();k++){
        Solide &partk = *(*listejoints(k)).part2;
        equa1 = &sys(2*k*listejoints.end()*2);
        equa2 = &sys((2*k+1)*listejoints.end()*2);
        secondterme = &constante(k*2);

        secondterme[0] += partk.masse*partk.acceleration;
        secondterme[1] += partk.getinertie()*partk.a_rot + (partk.w^partk.getinertie()*partk.w);
        for(i=0;i<partk.liens.end();i++){
            Joint &link = *partk.liens(i);
            if(link.is_pere(partk)){
                equa1[link.iforce*2] += Matr9d(1);
                equa2[link.iforce*2+1] += Matr9d(1);
                equa2[link.iforce*2] += (-link.AG1()).v();
            }else{
                equa1[link.iforce*2] += Matr9d(-1);
                equa2[link.iforce*2+1] += Matr9d(-1);
                equa2[link.iforce*2] += link.AG2().v();
            }
        }
    }
    gauss(sys, constante);
}
//*/

double Solveur::getenergie(){
    int k;
    double energie=0;
    for(k=0;k<listepartie.end();k++){
        Solide &pk = *listepartie(k);
        energie+=pk.masse*pk.v*pk.v;
        energie+=pk.w*(pk.getinertie()*pk.w);
        //energie+=2*pk.masse*(pk.p.y+1);// energie potentielle de pesanteur
    }
    energie*=0.5;
    return energie;
}

// erreur quadratique des contraintes
double Solveur::geterreur(){
    int k;
    double erreur = 0;
    for(k=0;k<listejoints.end();k++){erreur +=  (*listejoints(k)).errc * (*listejoints(k)).errc ;}
    return erreur;
}

/************** Corrections en position et vitesse ***************************/

void Solveur::corrigevars(){
    int k;
    for(k=0;k<listejoints.end();k++){
        (*listejoints(k)).updatevar();
        (*listejoints(k)).corrigevar();
    }
    //corrigepos();
}

double Solveur::corrigepos(){
    int k,i;

    // initialisations
    constante.setend(listejoints.end());
    for(k=0;k<listejoints.end();k++){constante(k)=Vect3d(); }
    sys.setend(listejoints.end()*listejoints.end());
    for(k=0;k<sys.end();k++)sys(k)=Matr9d();

    for(k=0;k<listejoints.end();k++){// précalculs des gradients
        (*listejoints(k)).updatevar();
    }
    for(k=0;k<listejoints.end();k++){// remplissage du système
        Joint &j = *listejoints(k);
        j.fillequa(&sys(listejoints.end()*k));
        constante(k) -= j.errc;// les termes constants s'ajoutent positivement à droite de l'equation
    }


    gauss(sys,constante);


    for(k=0;k<listepartie.end();k++){// calcul de l'accélération
        Solide &a = *listepartie(k);
        a.a_rot=Vect3d();
        a.acceleration=Vect3d();

        for(i=0;i<a.liens.end();i++){
            Joint &j = *a.liens(i);
            if(j.is_pere(a)){
                a.a_rot += transpo(j.Gq1)*constante(j.iforce);
                a.acceleration += transpo(j.Gp1)*constante(j.iforce);
            }else{
                a.a_rot += transpo(j.Gq2)*constante(j.iforce);
                a.acceleration += transpo(j.Gp2)*constante(j.iforce);
            }
        }

        a.a_rot=!a.getinertie()*a.a_rot;
        a.acceleration*=1./a.masse;
    }
    double coef=0;
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);
        coef = max(a.a_rot*a.a_rot, coef);
    }
    coef = min(1.,sqrt(4/coef));
    for(k=0;k<listepartie.end();k++){
        Solide &a = *listepartie(k);

        a.q = renorme(a.q + coef*.5*a.a_rot*a.q);
        a.p += coef*a.acceleration;
        a.R = a.q.matr();
    }
    //printf("coef: %lf",coef);


    // vide le système d'équation
    for(k=0;k<listejoints.end();k++){constante(k)=Vect3d(); }

    double erreur = 0;
    for(k=0;k<listejoints.end();k++){erreur +=  (*listejoints(k)).errc * (*listejoints(k)).errc ;}
    return erreur;

}

void Solveur::corrigev(){
    int k,i;

    // initialisations
    constante.setend(listejoints.end());
    for(k=0;k<listejoints.end();k++){constante(k)=Vect3d(); }
    sys.setend(listejoints.end()*listejoints.end());
    for(k=0;k<sys.end();k++)sys(k)=Matr9d();

    for(k=0;k<listejoints.end();k++){// précalculs des gradients
        (*listejoints(k)).updatevar();
    }
    for(k=0;k<listejoints.end();k++){// remplissage du système
        Joint &j = *listejoints(k);
        j.fillequa(&sys(listejoints.end()*k));
        constante(k) -= j.errv;// les termes constants s'ajoutent positivement à droite de l'equation
    }


    gauss(sys,constante);


    for(k=0;k<listepartie.end();k++){// calcul de l'accélération
        Solide &a = *listepartie(k);
        a.a_rot=Vect3d();
        a.acceleration=Vect3d();

        for(i=0;i<a.liens.end();i++){
            Joint &j = *a.liens(i);
            if(j.is_pere(a)){
                a.a_rot += transpo(j.Gq1)*constante(j.iforce);
                a.acceleration += transpo(j.Gp1)*constante(j.iforce);
            }else{
                a.a_rot += transpo(j.Gq2)*constante(j.iforce);
                a.acceleration += transpo(j.Gp2)*constante(j.iforce);
            }
        }

        a.a_rot=!a.getinertie()*a.a_rot;
        a.acceleration*=1./a.masse;

        a.w += a.a_rot;
        a.v += a.acceleration;
    }

}

/********************** Générateur d'impulsion ***************************/

/*
 Il s'agit de créer un joint qui fonctionnera pour un instant.
 On appliquera corrigevitesse, l'accélération êtant stockée,
 on pourra la réappliquée pour faire un choc plus ou moins mou.

 Au préalable il faut définir où se situera le choc

 Pour simplifier nous allons considérer des chocs générés par des liaisons rotule avec la terre.
 On défini ici un choc dès que le barycentre d'un solide atteint un certain rayon
 et que ce rayon scalaire la vitesse du barycentre du solide soit positif.
*/

void Solveur::impulsion(){
    int k;
    double rayon = 1.5;
    for(k=0;k<listepartie.end();k++){// détection d'un choc
        Solide &a = *listepartie(k);
        if(a.p*a.p  >  rayon*rayon && a.p*a.v > 0){

            Solide terre(0x1p+80, Vect3d(0x1p+23,0x1p+23,0x1p+23));// dimensions de la terre
            //Solide terre(0x1p+10, Vect3d(0x1p+5,0x1p+5,0x1p+5));
            Joint jplus(&a,&terre,Vect3d(0.,0,0),Vect3d(0.,0,0));// création de la contrainte de choc
            listejoints.add(&jplus);
            jplus.iforce=listejoints.end()-1;// ne pas oublier de donner le numero du joint au joint

            corrigev();

            double coef=1;
            //*
            for(k=0;k<listepartie.end();k++){// On double l'impusion pour faire un rebond
                Solide &b = *listepartie(k);
                b.w += coef*b.a_rot;
                b.v += coef*b.acceleration;
            }
            //*/

            listejoints.indice_suppr(listejoints.end()-1);// On supprime le joint
            a.liens.indice_suppr(a.liens.end()-1);// On supprime le lien entre le joint et le solide

            break;
        }
    }
}












